/*
 * =====================================================================================
 *
 *       Filename:  DepolarizerServer.h
 *
 *    Description:  DepolarizerServer
 *
 *        Version:  1.0
 *        Created:  14.06.2017 16:47:54
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
 *   Organization:  Budker Insitute of Nuclear Physics
 *
 * =====================================================================================
 */
#ifndef IBN_DEPOLARIZER_SERVER_H
#define IBN_DEPOLARIZER_SERVER_H

#include <map>
#include <list>
#include <functional>


#include "net/basic_server.h"
#include "barrier.h"

#include "device/Depolarizer.h"

class DepolarizerServer : public basic_server
{
  public:
    DepolarizerServer(int port=9090); //just DepolarizerServer withot any depolarizer (need to set up via set_depolarizer call)
    DepolarizerServer(int port, std::shared_ptr<device::Depolarizer> ); //DepolarizerServer with depolarizer device

    ~DepolarizerServer(void){};

    void set_depolarizer(std::shared_ptr<device::Depolarizer>); //set the depolarizer

  protected:

	template<typename ... Types> void print(std::ostream & os, int, const Types & ...  msg) const;
	template<typename ... Types> void print_error(const Types & ... msg) const ;
	template<typename ... Types> void print_warning(const Types & ... msg) const;
	template<typename ... Types> void print_info(const Types & ... msg) const;

    void handle_connection(std::shared_ptr<ip::tcp::socket>) final;

    std::shared_ptr<device::Depolarizer> fDepol;

	struct action_t 
	{
		std::function<void(double)> set; 
		std::function<double(void)> get;
	};

	std::map< int, action_t > set_get_actions;

	void setup_actions();

    void start_send_frequency(std::shared_ptr<ip::tcp::socket>  sock,  volatile bool & is_send);
    void send_frequency(std::shared_ptr<ip::tcp::socket>  sock, volatile bool & is_send);

    std::chrono::seconds send_period;
};

#endif
