/*
 * =====================================================================================
 *
 *       Filename:  barrier.h
 *
 *    Description:  this is realization of barrier
 *
 *        Version:  1.0
 *        Created:  22.06.2017 01:16:41
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
 *   Organization:  Budker Insitute of Nuclear Physics
 *
 * =====================================================================================
 */
#ifndef IBN_BARRIER_H
#define IBN_BARRIER_H

#include <condition_variable>
#include <atomic>


class mybarrier
{
  std::condition_variable cv; 
  std::condition_variable main_cv; 
  std::mutex  mutex;
  std::mutex  main_mutex;
  int nthread=0; //total number of thread
  int  nwait=0;  //number of waiting thread
  bool iswait = false; //wait flag
  public:
  void init(void)
  {
    if(nthread==0) return; //no thread in pool
    //std::unique_lock<std::mutex> lock(mutex);
    std::unique_lock<std::mutex> main_lock(main_mutex);
    iswait = true;
    while(nwait < nthread) main_cv.wait(main_lock);
    //while(nwait < nthread) cv.wait(lock);
  } 

  void release(void)
  {
    iswait = false;
    cv.notify_all(); //release all threads
  }

  //int & get_number(void) { return nthread; }


  class count_variable //this is the helper class which safely count variable
  {
    int & n;
    public:
    count_variable( count_variable &&  c ) : n (c.n) {}
    count_variable(int & i): n(i)
    {
      n++;
    }
    ~count_variable(void)
    {
      n--;
    }
  };
  count_variable  registrate(void)
  {
    return count_variable(nthread);
  }

  void wait(void)
  {
    if(nthread==0) return;
    if(iswait) 
    {
      count_variable c(nwait);
      if( nwait >= nthread )
      {
        //all sending thread come here last one notify first waiting thread which call init
        //cv.notify_one(); 
        //std::unique_lock<std::mutex> main_lock(main_mutex);
        main_cv.notify_all();
      }
      std::unique_lock<std::mutex> lock(mutex);
      while(iswait) cv.wait(lock);
    }
  }
};
#endif
