/*
 * =====================================================================================
 *
 *       Filename:  basic_server.h
 *
 *    Description:  Base server
 *
 *        Version:  1.0
 *        Created:  04.10.2017 17:53:41
 *       Revision:  none
 *       Compiler:  gcc
 *
 *         Author:  Ivan B. Nikolaev (ekherit), I.B.Nikolaev@inp.nsk.su
 *   Organization:  Budker Insitute of Nuclear Physics
 *
 * =====================================================================================
 */
#ifndef IBN_NET_BASIC_SERVER_H
#define IBN_NET_BASIC_SERVER_H

#include <memory>
#include <stdexcept>
#include <thread>
#include <iostream>

#include <boost/asio.hpp>
using  namespace boost::asio;


class  basic_server
{
  io_service service; //input output service
  std::shared_ptr<ip::tcp::endpoint> endpoint; //address
  std::shared_ptr<ip::tcp::acceptor> acceptor; //connection acceptor
  public:
  basic_server(int port=9090)
  {
    endpoint.reset(new ip::tcp::endpoint( ip::tcp::v4(), static_cast<short unsigned int>(port))); 
    acceptor.reset(new ip::tcp::acceptor(service, *endpoint));
  }

  virtual void run(void)
  {
    while(true)
    {
      try
      {
        std::shared_ptr<ip::tcp::socket> sock(new ip::tcp::socket(service)); //create the socket for a new connection
        acceptor->accept(*sock); //whait for the new connection
        std::thread t([this,sock](){handle_connection(sock); }); //handle the new connection in thread
        t.detach();
      }
      catch(std::exception & e)
      {
        std::cout << e.what() << std::endl;
      }
    }
  }

  virtual void handle_connection(std::shared_ptr<ip::tcp::socket> sock)
  {
    try 
    {
      char buf[1024];
      //size_t n;
      for(;;)
      {
        std::cout << "virtual basic_server::handle_connection" << std::endl;
        boost::system::error_code error;
        sock->read_some(buffer(buf),error);
        if (error == boost::asio::error::eof)
        {
          break; // Connection closed cleanly by peer.
        }
        else 
        {
          if (error) 
          {
            throw boost::system::system_error(error); // Some other error.
          }
        }
      }
    }
    catch(std::exception &e)
    {
      std::cerr << e.what() << std::endl;
    }
  }
};
#endif
